@@ -1,6 +1,6 @@
 /* -*- c-basic-offset: 2 -*- */
 /* gifsicle.c - gifsicle's main loop.
-   Copyright (C) 1997-2002 Eddie Kohler, kohler@icir.org
+   Copyright (C) 1997-2008 Eddie Kohler, kohler@icir.org
    This file is part of gifsicle.
 
    Gifsicle is free software. It is distributed under the GNU Public License,
@@ -19,7 +19,7 @@
 
 /* Need _setmode under MS-DOS, to set stdin/stdout to binary mode */
 /* Need _fsetmode under OS/2 for the same reason */
-#if defined(_MSDOS) || defined(_WIN32) || defined(__EMX__)
+#if defined(_MSDOS) || defined(_WIN32) || defined(__EMX__) || defined(__DJGPP__)
 # include <fcntl.h>
 # include <io.h>
 #endif
@@ -35,7 +35,8 @@ Gif_Stream *input = 0;
 const char *input_name = 0;
 static int unoptimizing = 0;
 
-int gif_read_flags = 0;
+static int gif_read_flags = 0;
+static int nextfile = 0;
 int gif_write_flags = 0;
 
 static int frames_done = 0;
@@ -181,22 +182,23 @@ static const char *output_option_types[] = {
 #define CROP_TRANSPARENCY_OPT	360
 #define CONSERVE_MEMORY_OPT	361
 #define MULTIFILE_OPT		362
+#define NEXTFILE_OPT		363
 
-#define LOOP_TYPE		(Clp_FirstUserType)
-#define DISPOSAL_TYPE		(Clp_FirstUserType + 1)
-#define DIMENSIONS_TYPE		(Clp_FirstUserType + 2)
-#define FRAME_SPEC_TYPE		(Clp_FirstUserType + 3)
-#define COLOR_TYPE		(Clp_FirstUserType + 4)
-#define POSITION_TYPE		(Clp_FirstUserType + 5)
-#define RECTANGLE_TYPE		(Clp_FirstUserType + 6)
-#define TWO_COLORS_TYPE		(Clp_FirstUserType + 7)
-#define COLORMAP_ALG_TYPE	(Clp_FirstUserType + 8)
-#define SCALE_FACTOR_TYPE	(Clp_FirstUserType + 9)
+#define LOOP_TYPE		(Clp_ValFirstUser)
+#define DISPOSAL_TYPE		(Clp_ValFirstUser + 1)
+#define DIMENSIONS_TYPE		(Clp_ValFirstUser + 2)
+#define FRAME_SPEC_TYPE		(Clp_ValFirstUser + 3)
+#define COLOR_TYPE		(Clp_ValFirstUser + 4)
+#define POSITION_TYPE		(Clp_ValFirstUser + 5)
+#define RECTANGLE_TYPE		(Clp_ValFirstUser + 6)
+#define TWO_COLORS_TYPE		(Clp_ValFirstUser + 7)
+#define COLORMAP_ALG_TYPE	(Clp_ValFirstUser + 8)
+#define SCALE_FACTOR_TYPE	(Clp_ValFirstUser + 9)
 
-Clp_Option options[] = {
+const Clp_Option options[] = {
   
   { "append", 0, APPEND_OPT, 0, 0 },
-  { "app-extension", 'x', APP_EXTENSION_OPT, Clp_ArgString, 0 },
+  { "app-extension", 'x', APP_EXTENSION_OPT, Clp_ValString, 0 },
   
   { "background", 'B', BACKGROUND_OPT, COLOR_TYPE, Clp_Negate },
   { "batch", 'b', 'b', 0, 0 },
@@ -206,16 +208,16 @@ Clp_Option options[] = {
   { "change-color", 0, CHANGE_COLOR_OPT, TWO_COLORS_TYPE, Clp_Negate },
   { "cinfo", 0, COLOR_INFO_OPT, 0, Clp_Negate },
   { "clip", 0, CROP_OPT, RECTANGLE_TYPE, Clp_Negate },
-  { "colors", 'k', COLORMAP_OPT, Clp_ArgInt, Clp_Negate },
+  { "colors", 'k', COLORMAP_OPT, Clp_ValInt, Clp_Negate },
   { "color-method", 0, COLORMAP_ALGORITHM_OPT, COLORMAP_ALG_TYPE, 0 },
   { "color-info", 0, COLOR_INFO_OPT, 0, Clp_Negate },
-  { "comment", 'c', COMMENT_OPT, Clp_ArgString, 0 },
-  { "no-comments", 'c', NO_COMMENTS_OPT, 0, Clp_OnlyNegated },
+  { "comment", 'c', COMMENT_OPT, Clp_ValString, 0 },
+  { "no-comments", 'c', NO_COMMENTS_OPT, 0, 0 },
   { "conserve-memory", 0, CONSERVE_MEMORY_OPT, 0, Clp_Negate },
   { "crop", 0, CROP_OPT, RECTANGLE_TYPE, Clp_Negate },
   { "crop-transparency", 0, CROP_TRANSPARENCY_OPT, 0, Clp_Negate },
   
-  { "delay", 'd', 'd', Clp_ArgInt, Clp_Negate },
+  { "delay", 'd', 'd', Clp_ValInt, Clp_Negate },
   { "delete", 0, DELETE_OPT, 0, 0 },
   { "disposal", 'D', DISPOSAL_OPT, DISPOSAL_TYPE, Clp_Negate },
   { "dither", 'f', DITHER_OPT, 0, Clp_Negate },
@@ -223,7 +225,7 @@ Clp_Option options[] = {
   
   { "explode", 'e', 'e', 0, 0 },
   { "explode-by-name", 'E', 'E', 0, 0 },
-  { "extension", 0, EXTENSION_OPT, Clp_ArgString, 0 },
+  { "extension", 0, EXTENSION_OPT, Clp_ValString, 0 },
   { "no-extensions", 'x', NO_EXTENSIONS_OPT, 0, 0 },
   { "extension-info", 0, EXTENSION_INFO_OPT, 0, Clp_Negate },
   
@@ -244,18 +246,19 @@ Clp_Option options[] = {
   { "method", 0, COLORMAP_ALGORITHM_OPT, COLORMAP_ALG_TYPE, 0 },
   { "multifile", 0, MULTIFILE_OPT, 0, Clp_Negate },
   
-  { "name", 'n', NAME_OPT, Clp_ArgString, 0 },
-  { "no-names", 'n', NO_NAME_OPT, 0, Clp_OnlyNegated },
+  { "name", 'n', NAME_OPT, Clp_ValString, 0 },
+  { "nextfile", 0, NEXTFILE_OPT, 0, Clp_Negate },
+  { "no-names", 'n', NO_NAME_OPT, 0, 0 },
   
-  { "optimize", 'O', OPTIMIZE_OPT, Clp_ArgInt, Clp_Negate | Clp_Optional },
-  { "output", 'o', OUTPUT_OPT, Clp_ArgStringNotOption, 0 },
+  { "optimize", 'O', OPTIMIZE_OPT, Clp_ValInt, Clp_Negate | Clp_Optional },
+  { "output", 'o', OUTPUT_OPT, Clp_ValStringNotOption, 0 },
   
   { "position", 'p', POSITION_OPT, POSITION_TYPE, Clp_Negate },
   
   { "replace", 0, REPLACE_OPT, FRAME_SPEC_TYPE, 0 },
   { "resize", 0, RESIZE_OPT, DIMENSIONS_TYPE, Clp_Negate },
-  { "resize-width", 0, RESIZE_WIDTH_OPT, Clp_ArgUnsigned, Clp_Negate },
-  { "resize-height", 0, RESIZE_HEIGHT_OPT, Clp_ArgUnsigned, Clp_Negate },
+  { "resize-width", 0, RESIZE_WIDTH_OPT, Clp_ValUnsigned, Clp_Negate },
+  { "resize-height", 0, RESIZE_HEIGHT_OPT, Clp_ValUnsigned, Clp_Negate },
   { "resiz", 0, RESIZE_OPT, DIMENSIONS_TYPE, Clp_Negate },
   { "resi", 0, RESIZE_OPT, DIMENSIONS_TYPE, Clp_Negate },
   { "res", 0, RESIZE_OPT, DIMENSIONS_TYPE, Clp_Negate },
@@ -282,12 +285,12 @@ Clp_Option options[] = {
   { "same-screen", 0, SAME_LOGICAL_SCREEN_OPT, 0, 0 },
   { "same-transparent", 0, SAME_TRANSPARENT_OPT, 0, 0 },
   
-  { "transform-colormap", 0, COLOR_TRANSFORM_OPT, Clp_ArgStringNotOption,
+  { "transform-colormap", 0, COLOR_TRANSFORM_OPT, Clp_ValStringNotOption,
     Clp_Negate },
   { "transparent", 't', 't', COLOR_TYPE, Clp_Negate },
   
   { "unoptimize", 'U', UNOPTIMIZE_OPT, 0, Clp_Negate },
-  { "use-colormap", 0, USE_COLORMAP_OPT, Clp_ArgString, Clp_Negate },
+  { "use-colormap", 0, USE_COLORMAP_OPT, Clp_ValString, Clp_Negate },
   
   { "verbose", 'V', VERBOSE_OPT, 0, Clp_Negate },
   { 0, 'v', VERBOSE_OPT, 0, Clp_Negate },
@@ -471,6 +474,84 @@ gifread_error(const char *message, int which_image, void *thunk)
   }
 }
 
+struct StoredFile {
+  FILE *f;
+  struct StoredFile *next;
+  char name[1];
+};
+
+static struct StoredFile *stored_files = 0;
+
+static FILE *
+open_giffile(const char *name)
+{
+  struct StoredFile *sf;
+  FILE *f;
+  
+  if (name == 0 || strcmp(name, "-") == 0) {
+#ifndef OUTPUT_GIF_TO_TERMINAL
+    extern int isatty(int);
+    if (isatty(fileno(stdin))) {
+      error("<stdin>: is a terminal");
+      return NULL;
+    }
+#endif
+#if defined(_MSDOS) || defined(_WIN32)
+    _setmode(_fileno(stdin), _O_BINARY);
+#elif defined(__DJGPP__)
+    setmode(fileno(stdin), O_BINARY);
+#elif defined(__EMX__)
+    _fsetmode(stdin, "b");
+#endif
+    return stdin;
+  }
+    
+  if (nextfile)
+    for (sf = stored_files; sf; sf = sf->next)
+      if (strcmp(name, sf->name) == 0)
+	return sf->f;
+  
+  f = fopen(name, "rb");
+  
+  if (f && nextfile) {
+    sf = (struct StoredFile *) malloc(sizeof(struct StoredFile) + strlen(name));
+    sf->f = f;
+    sf->next = stored_files;
+    stored_files = sf;
+    strcpy(sf->name, name);
+  } else if (!f)
+    error("%s: %s", name, strerror(errno));
+  
+  return f;
+}
+
+static void
+close_giffile(FILE *f, int final)
+{
+  struct StoredFile **sf_pprev, *sf;
+  
+  if (!final && nextfile) {
+    int c = getc(f);
+    if (c == EOF)
+      final = 1;
+    else
+      ungetc(c, f);
+  }
+
+  for (sf_pprev = &stored_files; (sf = *sf_pprev); sf_pprev = &sf->next)
+    if (sf->f == f) {
+      if (final) {
+	fclose(f);
+	*sf_pprev = sf->next;
+	free((void *) sf);
+      }
+      return;
+    }
+
+  if (f != stdin)
+    fclose(f);
+}
+
 void
 input_stream(const char *name)
 {
@@ -494,21 +575,13 @@ input_stream(const char *name)
 
   if (infoing == 1)
     set_mode(INFOING);
-  
-  if (name == 0 || strcmp(name, "-") == 0) {
-#if defined(_MSDOS) || defined(_WIN32)
-    _setmode(_fileno(stdin), _O_BINARY);
-#elif defined(__EMX__)
-    _fsetmode(stdin, "b");
-#endif
-    f = stdin;
+
+  f = open_giffile(name);
+  if (!f)
+    return;
+  if (f == stdin) {
     name = "<stdin>";
     input_name = 0;
-  } else
-    f = fopen(name, "rb");
-  if (!f) {
-    error("%s: %s", name, strerror(errno));
-    return;
   }
   main_name = name;
 
@@ -527,8 +600,9 @@ input_stream(const char *name)
   if (i == EOF) {
     if (!(gif_read_flags & GIF_READ_TRAILING_GARBAGE_OK))
       error("%s: empty file", name);
-    if (f != stdin)
-      fclose(f);
+    else if (nextfile)
+      error("%s: no more images in file", name);
+    close_giffile(f, 1);
     return;
   }
   ungetc(i, f);
@@ -548,9 +622,9 @@ input_stream(const char *name)
     else
       error("%s: trailing garbage ignored", main_name);
     Gif_DeleteStream(gfs);
-    if (verbosing) verbose_close('>');
-    if (f != stdin)
-      fclose(f);
+    if (verbosing)
+      verbose_close('>');
+    close_giffile(f, 1);
     return;
   }
 
@@ -634,10 +708,9 @@ input_stream(const char *name)
   gfs->refcount++;
 
   /* Read more files. */
-  if (gif_read_flags & GIF_READ_TRAILING_GARBAGE_OK)
+  if ((gif_read_flags & GIF_READ_TRAILING_GARBAGE_OK) && !nextfile)
     goto retry_file;
-  else if (f != stdin)
-    fclose(f);
+  close_giffile(f, 0);
 }
 
 void
@@ -772,12 +845,14 @@ write_stream(const char *output_name, Gif_Stream *gfs)
 #ifndef OUTPUT_GIF_TO_TERMINAL
     extern int isatty(int);
     if (isatty(fileno(stdout))) {
-      error("not writing to <stdout>: it's a terminal");
+      error("<stdout>: is a terminal");
       return;
     }
 #endif
 #if defined(_MSDOS) || defined(_WIN32)
     _setmode(_fileno(stdout), _O_BINARY);
+#elif defined(__DJGPP__)
+    setmode(fileno(stdout), O_BINARY);
 #elif defined(__EMX__)
     _fsetmode(stdout, "b");
 #endif
@@ -975,7 +1050,7 @@ static int
 handle_extension(Clp_Parser *clp, int is_app)
 {
   Gif_Extension *gfex;
-  const char *extension_type = clp->arg;
+  const char *extension_type = clp->vstr;
   const char *extension_body = Clp_Shift(clp, 1);
   if (!extension_body) {
     Clp_OptionError(clp, "%O requires two arguments");
@@ -1291,7 +1366,7 @@ main(int argc, char *argv[])
      case NAME_OPT:
       if (clp->negated) goto no_names;
       MARK_CH(frame, CH_NAME);
-      def_frame.name = clp->arg;
+      def_frame.name = clp->vstr;
       break;
       
      no_names:
@@ -1310,7 +1385,7 @@ main(int argc, char *argv[])
       if (clp->negated) goto no_comments;
       MARK_CH(frame, CH_COMMENT);
       if (!def_frame.comment) def_frame.comment = Gif_NewComment();
-      Gif_AddComment(def_frame.comment, clp->arg, -1);
+      Gif_AddComment(def_frame.comment, clp->vstr, -1);
       break;
       
      no_comments:
@@ -1506,7 +1581,7 @@ main(int argc, char *argv[])
       if (clp->negated)
 	def_output_data.loopcount = -1;
       else
-	def_output_data.loopcount = (clp->have_arg ? clp->val.i : 0);
+	def_output_data.loopcount = (clp->have_val ? clp->val.i : 0);
       break;
       
      case SAME_LOOPCOUNT_OPT:
@@ -1519,7 +1594,7 @@ main(int argc, char *argv[])
       if (clp->negated)
 	def_output_data.optimizing = 0;
       else
-	def_output_data.optimizing = (clp->have_arg ? clp->val.i : 1);
+	def_output_data.optimizing = (clp->have_val ? clp->val.i : 1);
       break;
       
      case UNOPTIMIZE_OPT:
@@ -1531,11 +1606,9 @@ main(int argc, char *argv[])
 
      case CAREFUL_OPT: {
        if (clp->negated)
-	 gif_read_flags = gif_write_flags = 0;
-       else {
-	 gif_read_flags = 0;
+	 gif_write_flags = 0;
+       else
 	 gif_write_flags = GIF_WRITE_CAREFUL_MIN_CODE_SIZE;
-       }
        break;
      }
      
@@ -1559,7 +1632,7 @@ main(int argc, char *argv[])
 	  (output_transforms, &pipe_color_transformer);
       else
 	output_transforms = append_color_transform
-	  (output_transforms, &pipe_color_transformer, (void *)clp->arg);
+	  (output_transforms, &pipe_color_transformer, (void *)clp->vstr);
       break;
       
      case COLORMAP_OPT:
@@ -1582,7 +1655,7 @@ main(int argc, char *argv[])
       if (clp->negated)
 	def_output_data.colormap_fixed = 0;
       else
-	set_new_fixed_colormap(clp->arg);
+	set_new_fixed_colormap(clp->vstr);
       break;
       
      case COLORMAP_ALGORITHM_OPT:
@@ -1669,8 +1742,19 @@ main(int argc, char *argv[])
      case MULTIFILE_OPT:
       if (clp->negated)
 	gif_read_flags &= ~GIF_READ_TRAILING_GARBAGE_OK;
-      else
+      else {
 	gif_read_flags |= GIF_READ_TRAILING_GARBAGE_OK;
+	nextfile = 0;
+      }
+      break;
+
+     case NEXTFILE_OPT:
+      if (clp->negated)
+        gif_read_flags &= ~GIF_READ_TRAILING_GARBAGE_OK;
+      else {
+	gif_read_flags |= GIF_READ_TRAILING_GARBAGE_OK;
+	nextfile = 1;
+      }
       break;
 
      case VERSION_OPT:
@@ -1679,7 +1763,7 @@ main(int argc, char *argv[])
 #else
       printf("LCDF Gifsicle %s\n", VERSION);
 #endif
-      printf("Copyright (C) 1997-2005 Eddie Kohler\n\
+      printf("Copyright (C) 1997-2008 Eddie Kohler\n\
 This is free software; see the source for copying conditions.\n\
 There is NO warranty, not even for merchantability or fitness for a\n\
 particular purpose.\n");
@@ -1693,18 +1777,18 @@ particular purpose.\n");
       
      case OUTPUT_OPT:
       MARK_CH(output, CH_OUTPUT);
-      if (strcmp(clp->arg, "-") == 0)
+      if (strcmp(clp->vstr, "-") == 0)
 	def_output_data.output_name = 0;
       else
-	def_output_data.output_name = clp->arg;
+	def_output_data.output_name = clp->vstr;
       break;
       
       /* NONOPTIONS */
       
      case Clp_NotOption:
-      if (clp->arg[0] != '#' || !frame_argument(clp, clp->arg)) {
+      if (clp->vstr[0] != '#' || !frame_argument(clp, clp->vstr)) {
 	input_done();
-	input_stream(clp->arg);
+	input_stream(clp->vstr);
       }
       break;
       
